{% extends "!layout.html" %}
{% block extrahead %}
  <link href="{{ pathto("_static/style.css", True) }}" rel="stylesheet" type="text/css">
{% endblock %}

{% block sidebartitle %}
  <div class="switch-menus">
    <div class="version">
      <strong>Version:</strong> {{ version }}
    </div>
  </div>

  {{ super() }}
{% endblock %}
